from django.db import models
from django.core.exceptions import ValidationError
from django.db.models import Q
from simple_history.models import HistoricalRecords

class ModelBase(models.Model):
    class Meta:
        abstract = True
    history = HistoricalRecords(related_name='log', inherit=True)

    @classmethod
    def get_field_names(cls):
        ''' Return a list of the db field names for objects of this class '''
        return [f.name for f in cls._meta.get_fields()]

class CheckOverlapMixin:
    ''' 
        This mixin adds cleaning functionality to a model that will not allow a range overlap defined by two attributes.
        default behavior is to include the boundaries, but can be overwritten
    '''
    start_attr = '' # Name of the attr that starts the range (must be overwritten)
    end_attr = '' # Name of the attr that ends the range (must be overwritten)
    extra_filter_attrs = [] # Additional filter attrs to apply before checking for overlap
    include_end_boundary = True
    include_start_boundary = True
    clean_during_save = True

    def clean(self):
        super().clean()
        self.check_overlap()

    def save(self, *args, **kwargs):
        if self.clean_during_save:
            self.clean()
        super().save(*args,**kwargs)

    def check_overlap(self):
        # Overlapping ranges will always meet BOTH of the following conditions
        # 1.) existing end date is >= new start date
        # 2.) existing start date is <= new end date
        fkwargs= {} # initialize filter kwargs
        #  Get the look ups based on boundary cond settings
        slu = 'lte' if self.include_start_boundary else 'lt'
        elu = 'gte' if self.include_end_boundary else 'gt'
        fkwargs[f'{self.end_attr}__{elu}'] = getattr(self, self.start_attr, None)  # condition 1
        fkwargs[f'{self.start_attr}__{slu}'] = getattr(self, self.end_attr, None)  # condition 2
        for fattr in self.extra_filter_attrs:
            fkwargs[fattr] = getattr(self, fattr, None)
        # Count the number of objects (other than the object itself) that meet both conditions
        try:
            bad_count = self.__class__.objects.filter(
                ~Q(id=self.pk),
            ).filter(**fkwargs).count()
        except Exception as e:
            raise ValidationError('Check overlap failed')
        #No other objects should have met that criteria
        if bad_count > 0:
            msg = f'{self.__class__.__name__} cannot overlap another instance'
            if self.extra_filter_attrs != []:
                msg = msg + f' with the same {self.extra_filter_attrs}'
            raise ValidationError(
                msg
            )


from crispy_forms.helper import FormHelper
from crispy_forms.layout import *
from crispy_forms.bootstrap import *
from django import forms

class ModelFormBase(forms.ModelForm):
    DISABLE = []
    LIMIT_QS = []
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.DISABLE:
            self.fields[field].disabled = True
        instance = kwargs.get('instance')
        if instance:
            for field in self.LIMIT_QS:
                qs = self.fields[field].queryset
                obj = getattr(instance,field,None)
                if obj:
                    self.fields[field].queryset = qs.filter(id=obj.pk)
                else:
                    self.fields[field].queryset = qs.none()
        self.helper = FormHelper()
        self.helper.add_input(Submit('submit', 'Submit'))
        self.helper.add_input(
            Button(
                'cancel', 'Cancel', css_class='btn-primary',
                onclick="history.back();"
            )
        )


from django.urls import path
from django.contrib.auth.views import LoginView, LogoutView

urlpatterns = [
    path('login/', LoginView.as_view(template_name='django_aux/login.html'), name='login'),
    path('logout/', LogoutView.as_view(template_name='django_aux/logout.html'), name='logout'),
]



from django_filters import *
import django.db.models as models
from bootstrap_datepicker_plus.widgets import DatePickerInput, DateTimePickerInput
from crispy_forms.helper import FormHelper
from django import forms
from crispy_forms.layout import *


class MetaBase:
    filter_overrides = {
        models.DateTimeField: {
            'filter_class': DateTimeFilter,
            'extra': lambda f: {
                'widget': DateTimePickerInput,
            },
        },
        models.DateField: {
            'filter_class': DateFilter,
            'extra': lambda f: {
                'widget': DatePickerInput,
            },
        },
        models.BooleanField: {
            'filter_class': MultipleChoiceFilter,
            'extra': lambda f: {
                'widget': forms.CheckboxSelectMultiple,
                'choices': ((True, 'Yes'), (False, 'No'))
            },
        },
    }

class FilterSetBase(FilterSet):
    ''' A BaseFilter class that initializes a crispy form helper 
    with submit and clear filter buttons '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.helper = FormHelper()
        self.form.helper.disable_csrf = True
        self.form.helper.add_input(
            Submit('submit', 'Apply Filter')
        )
        self.form.helper.add_input(
            Submit('clear_filter', 'Clear Filter', css_class='btn-warning')
        )


class PlotSettingsFilterMixin(FilterSet):
    ''' The purpose of this filter is to add in a standard form for plot
    settings to a normal model filter'''

    def filter_queryset(self, queryset):
        qdict = self.form.cleaned_data.copy()
        for key in ['x', 'y', 'plot_type', 'color', 'aggregate_by', 'N_min', 'y_min', 'y_max']:
            if key in qdict.keys():
                qdict.pop(key)
        for name, value in qdict.items():
            if value in ['', None, []]:
                continue
            queryset = self.filters[name].filter(queryset, value)
            assert isinstance(queryset, models.QuerySet), \
                "Expected '%s.%s' to return a QuerySet, but got a %s instead." \
                % (type(self).__name__, name, type(queryset).__name__)
        return queryset

    def __init__(self, *args, choices=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.form.fields['x'] = forms.ChoiceField(choices=choices.get('X_CHOICES'))
        self.form.fields['y'] = forms.ChoiceField(choices=choices.get('Y_CHOICES'))
        self.form.fields['plot_type'] = forms.ChoiceField(choices=choices.get('PLOT_TYPE_CHOICES'))
        self.form.fields['color'] = forms.ChoiceField(choices=choices.get('COLOR_CHOICES'), required=False)
        self.form.fields['aggregate_by'] = forms.ChoiceField(choices=choices.get('AGG_CHOICES'), required=False)
        self.form.fields['N_min'] = forms.IntegerField(initial=0, required=False, label='Sample Size Min')
        self.form.fields['y_min'] = forms.FloatField(required=False, label='Y Min Value')
        self.form.fields['y_max'] = forms.FloatField(required=False, label='Y Max Value')
        self.extra_layout = Layout(
            Fieldset('Plot Settings',
                Row(
                    Div('aggregate_by', css_class='ml-2 col-flex'),
                    Div('x', css_class='ml-2 col-flex'),
                    Div('y', css_class='ml-2 col-flex'),
                    Div('color', css_class='ml-2 col-flex'),
                    Div('plot_type', css_class='ml-2 col-flex'),
                    Div('N_min', css_class='ml-2 col-flex',style='width:125px'),
                    Div('y_min', css_class='ml-2 col-flex',style='width:125px'),
                    Div('y_max', css_class='ml-2 col-flex',style='width:125px'),
                ),
            ),
        )


